/**
* client_rdb.c - Receipt storage at the client side
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "client_rdb.h"

//max rcpt size (using rsa208 keys)
int size_of_rcpt = sizeof(unsigned char)  + 2*sizeof(unsigned short) + 2*sizeof(unsigned long) + 5*256;

//RB tree CALLBACKS
//returns the key for an RBTNODE storing receipts (node->data is a PWNC_RCPT*)
void *rcpt_key_func(RBTNODE *node) {
	if (node == NULL_NODE) {
		return 0;
	}
	return (void*)((RBTREE_RCPT_CL*)node->data)->rcpt->timestamp;
}

//sets key value to an RBTNODE storing receipts
void rcpt_set_key(RBTNODE *node, void *keyval) {
	if (node == NULL_NODE) {
		return;
	}
	((RBTREE_RCPT_CL*)node->data)->rcpt->timestamp = (unsigned long)keyval;
}

//compares (the keys of) two RBTNODEs storing receipts
int rcpt_comp(RBTNODE *node1, RBTNODE *node2) {
	if (node1 == NULL_NODE) {
		if (node2 == NULL_NODE) {
			return 0;
		}
		else {
			return -1;
		}
	}
	else
	if (node2 == NULL_NODE) {
		return 1;
	}

	if ( ((RBTREE_RCPT_CL*)node1->data)->rcpt->timestamp > ((RBTREE_RCPT_CL*)node2->data)->rcpt->timestamp) {
		return 1;
	}
	if ( ((RBTREE_RCPT_CL*)node1->data)->rcpt->timestamp < ((RBTREE_RCPT_CL*)node2->data)->rcpt->timestamp) {
		return -1;
	}
	if ( ((RBTREE_RCPT_CL*)node1->data)->rcpt->timestamp == ((RBTREE_RCPT_CL*)node2->data)->rcpt->timestamp) {
		return 0;
	}
	return 0;
}

//constructs a new RBTNODE and returns it
//the new node contains the rcpt in the data field (that is an RBTREE_RCPT object containing rcpt)
RBTNODE *RBTNODE_new(PWNC_RCPT *rcpt, const void *key_func, const void *set_key_func, const void *comp_func) {
	RBTNODE *retval = (RBTNODE*)malloc(sizeof(RBTNODE));
	RBTREE_RCPT_CL *ndata = (RBTREE_RCPT_CL*)malloc(sizeof(RBTREE_RCPT_CL));

	memset(retval, 0, sizeof(RBTNODE));
	memset(ndata, 0, sizeof(RBTREE_RCPT_CL));

	ndata->rcpt = rcpt;
	ndata->disk_pos = 0;

	retval->data = (void*)ndata;
	retval->left = NULL_NODE;
	retval->right = NULL_NODE;
	retval->parent = NULL_NODE;
	retval->key_func = key_func;
	retval->comp_func = comp_func;
	retval->set_key_func = set_key_func;
	return retval;
}

//returns the timestamp of the most recent receipt
unsigned long get_most_recent_ts() {
	RBTNODE *cur = rcpt_tree;
	while (cur->right != NULL_NODE) {
		cur = cur->right;
	}
	if (cur == NULL_NODE)
		return 0;

	return (unsigned long)cur->key_func(cur);
}


#define FL_ERROR -1
#define FL_EMPTY_FILE -2
//finds a free position where a new receipt can be stored (in perm storage)
//if -1 is returned, either an error has occured or no space is left (threshold reached) so the receipt cannot be stored
long pop_free_pos() {
	FILE *fp = NULL;
	int fd = 0;

	int retval = 0;
	long start_of_file, end_of_file, file_size;

	fd = open(RECEIPTS_DB_FREE_LIST, O_RDWR);
	if (fd == -1) {
		printf("Error opening free list file\n");
		return (long)FL_ERROR;
	}

	fp = fdopen(fd, "r+");
	if (!fp) {
		printf("Error opening free list file\n");
		close(fd);
		return (long)FL_ERROR;
	}

	end_of_file = ftell(fp);
	fseek(fp, 0L, SEEK_SET);
	start_of_file = ftell(fp);
	file_size = end_of_file - start_of_file;

	if (file_size == 0) {
		//empty file <==> no free positions
		return (long)FL_EMPTY_FILE;
	}

	//if file is not empty -> move to the last record
	if (fseek(fp, -1L, SEEK_END) == -1) {
		printf("Failed to get last record\n");
		return (long)FL_ERROR;
	}

	//read the free pos
	if (fread(&retval, sizeof(long), 1, fp) != sizeof(long)) {
		printf("Failed to read record\n");
		return (long)FL_ERROR;
	}

	//remove last record - truncate file
	ftruncate( fd, end_of_file - start_of_file - sizeof(long) );

	fclose(fp);

	return retval;
}

//writes a new free pos at the end of the free-pos list
int push_free_pos(long free_pos) {
	FILE *fp = NULL;

	//check if free_pos is outside the valid range
	if ( (free_pos < 0) && (free_pos > RECEIPT_THRESHOLD) ) {
		printf("Invalid free_pos value\n");
		return FL_ERROR;
	}

	fp = fopen(RECEIPTS_DB_FREE_LIST, "r+");
	if (!fp) {
		printf("Error opening free list file\n");
		return FL_ERROR;
	}

	//move past the last record
	fseek(fp, 0L, SEEK_END);

	fwrite(&free_pos, sizeof(long), 1, fp);
	return 0;
}

//gets a list of the free positions in the db
//On error ---> num_of_items = -1
//If the file is empty ---> num_of_items = 0
//(In both above cases NULL is returned, else an array of longs is returned)
long *get_free_list(int *num_of_items) {
	FILE *fp = NULL;
	long end_of_file, start_of_file, file_size;
	long *retval = NULL;
	int i = 0;

	fp = fopen(RECEIPTS_DB_FREE_LIST, "r");
	if (!fp) {
		printf("Error opening free list file\n");
		*num_of_items = -1;
		return NULL;
	}

	end_of_file = ftell(fp);
	fseek(fp, 0L, SEEK_SET);
	start_of_file = ftell(fp);
	file_size = end_of_file - start_of_file;

	//if file is empty --> set num_of_items to 0 and return NULL
	if (file_size == 0) {
		*num_of_items = 0;
		fclose(fp);
		return NULL;
	}

	retval = (long*)malloc( file_size/sizeof(long) );
	memset(retval, 0, file_size/sizeof(long) );

	for (i=0; i<file_size/sizeof(long); i++) {
		//read free list (it stores long integers which represent free positions in the db)
		if ( fread(&(retval[i]), 1, sizeof(long), fp) != sizeof(long) ) {
			SAFE_FREE(retval);
			*num_of_items = -1;
			fclose(fp);
			return NULL;
		}
	}
	fclose(fp);
	*num_of_items = file_size/sizeof(long);
	return retval;
}

int is_free_pos(long *free_list, int list_len, long pos) {
	int i = 0;
	for (i=0; i<list_len; i++) {
		if (free_list[i] == pos) {
			return 1;
		}
	}
	return 0;
}


//Opens the receipt db and loads the receipts in memory
//this func reads the receipts from the db into a red black tree
void load_receipts() {
	long *free_positions = NULL; //list of free positions
	int flist_size = 0; //number of free positions as read by the file
	FILE *fp = NULL;
	long offset = 0; //receipt position in the file (measured in bytes)
	int pos = 0; //receipt position in the file (measured in records)
	PWNC_RCPT *rcpt = NULL;
	RBTNODE *new_tree_node = NULL_NODE;

	free_positions = get_free_list(&flist_size);
	if ( (free_positions == NULL) && (flist_size == -1) ){
		//error retrieving free list
		printf("Error retrieving list of free positions\n");
		return;
	}

	//opens receipts file
	fp = fopen(RECEIPTS_DB_FILE, "r");
	if (!fp) {
		printf("Error opening rdb file\n");
		SAFE_FREE(free_positions);
		return;
	}

	//if flist is NULL, do not check for free positions, just load the receipt read
	while (!feof(fp)) {

		//if the position is free, we skip it
		if (free_positions && is_free_pos(free_positions, flist_size, pos)) {
			offset += size_of_rcpt;
			pos++;
			continue;
		}
		else {
			//receipt is valid
			//read it
			rcpt = read_rcpt_file(fp, offset);
			//insert it into tree and hash table and adjust teams
			if (rcpt) {
				new_tree_node = RBTNODE_new(rcpt, rcpt_key_func, rcpt_set_key, rcpt_comp);
				if (num_of_receipts_in_tree == 0) {
					//if the tree is empty (once in a lifetime...)
					rcpt_tree = new_tree_node;
				}
				else {
					rbtree_insert(&rcpt_tree, new_tree_node );
				}
				num_of_receipts_in_tree++;
			}
			offset += size_of_rcpt;
			pos++;
		}
	}
	fclose(fp);
}



int handle_receipt_storage(RBTREE_RCPT_CL *to_be_handled, int action) {
	FILE *fp = NULL;
	int pos = 0;
	int final_pos = 0;

	fp = fopen(RECEIPTS_DB_FILE, "r+");
	if (!fp) {
		printf("Error opening rdb file\n");
		return -1;
	}
//	printf("Opened %s \n", RECEIPTS_DB_FILE);
	switch (action) {
		case RDB_ACTION_INSERT:
			pos = pop_free_pos();
			if (pos == (long)FL_ERROR) {
				printf("Error getting free pos\n");
				return -1;
			}

			if (pos == (long)FL_EMPTY_FILE) {
				if (num_of_receipts_in_tree != RECEIPT_THRESHOLD) {
					final_pos = num_of_receipts_in_tree;
				}
				else {
					//cannot store receipt because db is full
					return -1;
				}
			}
			else {
				//else: the new receipt will overwrite a deleted one (will be written on pos)
				final_pos = pos;
			}

			//insert receipt in a free pos in the file
			if (io_store_receipt( to_be_handled->rcpt, fp, final_pos*size_of_rcpt) == -1) {
				//store receipt failed, so we can push the free pos back into the file and return
				if (pos != (long)FL_EMPTY_FILE) {
					push_free_pos(pos);
				}
				return -1;
			}
			//printf("Stored a receipt in %d\n\n", final_pos);
			//set node's disk_pos
			to_be_handled->disk_pos = final_pos;
		break;

		case RDB_ACTION_REMOVE:
			//in case of a deletion, we just add the pos that the receipt occupied in the list of free receipts
			push_free_pos( to_be_handled->disk_pos );
		break;

		default:
			return -1;
		break;
	} //endsw
	return 0;
}

//inserts a receipts in the tree and in the database
//we suppose that the check whether there is space for this receipt has ALREADY been performed
int insert_receipt(PWNC_RCPT *rcpt) {
	RBTNODE *new_tree_node = NULL_NODE;
	if (num_of_receipts_in_tree >= RECEIPT_THRESHOLD) {
		return -1;
	}

	if (!rcpt) {
		return -1;
	}

	new_tree_node = RBTNODE_new(rcpt, rcpt_key_func, rcpt_set_key, rcpt_comp);
	if (num_of_receipts_in_tree == 0) {
		//if the tree is empty (once in a lifetime...)
		rcpt_tree = new_tree_node;
	}
	else {
		rbtree_insert(&rcpt_tree, new_tree_node );
	}
	num_of_receipts_in_tree++;
	//insert into db
	handle_receipt_storage( (RBTREE_RCPT_CL*)new_tree_node->data, RDB_ACTION_INSERT);
	return 0;
}

RBTREE_RCPT_CL* RBTREE_RCPT_CL_new(PWNC_RCPT *rcpt) {
	RBTREE_RCPT_CL *retval = (RBTREE_RCPT_CL*)malloc(sizeof(RBTREE_RCPT_CL));
	memset(retval, 0, sizeof(RBTREE_RCPT_CL));
	retval->rcpt = rcpt;
	retval->disk_pos = -1;
	return retval;
}

void client_rdb_range_query( void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, unsigned long timestamp ) {
	PWNC_RCPT *testrcpt = (PWNC_RCPT*)malloc(sizeof(PWNC_RCPT));
	RBTREE_RCPT_CL *testdata = RBTREE_RCPT_CL_new(testrcpt);
	RBTNODE testnode  = {testdata, NULL_NODE, NULL_NODE, NULL_NODE, NODE_COLOR_BLACK, (void*)&rcpt_key_func, (void*)&rcpt_comp, (void*)&rcpt_set_key};
	testnode.set_key_func(&testnode, (void*)timestamp);

	//perform a range query on the tree
	rbtree_range_query(rcpt_tree, do_stuff, do_stuff_args, &testnode);

	//finally, free memory
	SAFE_FREE(testrcpt);
	SAFE_FREE(testdata);
}
