/**
* rbtree.h - Red-Black tree data structure implementation
* (internal <--> network (ascii) representation, message generation/parsing, etc)
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/**
RED-BLACK tree data structure implementation
see CLRS chapter 13
Each tree node holds:
- Color (r/b)
- Receipt (node key: receipt timestamp) list (receipts with equal timestamp)
- left/right/parent pointers
*/

#ifndef _PWNC_RBTREE_H
#define _PWNC_RBTREE_H

#include "io.h"

#ifndef RECEIPT_THRESHOLD
#define RECEIPT_THRESHOLD 5000
#endif

#define NODE_COLOR_RED 1
#define NODE_COLOR_BLACK 2

typedef struct rbtree_node_t {
	void *data;
	struct rbtree_node_t *left, *right, *parent;
	unsigned short color;
	void* (*key_func)(struct rbtree_node_t *);
	int (*comp_func)(const struct rbtree_node_t *, const struct rbtree_node_t *);
	void (*set_key_func)(struct rbtree_node_t *, void *);
} RBTNODE;

RBTNODE *NULL_NODE;

void tree_init();

void left_rotate(RBTNODE **tree, RBTNODE *x);
void right_rotate(RBTNODE **tree, RBTNODE *y);
void rbtree_insert_fixup(RBTNODE **tree, RBTNODE *z);
void rbtree_insert(RBTNODE **tree, RBTNODE *z);
RBTNODE *tree_minimum(RBTNODE *x);
RBTNODE *tree_successor(RBTNODE *x);
void rbtree_delete_fixup(RBTNODE **tree, RBTNODE *x);
RBTNODE *rbtree_delete(RBTNODE **tree, RBTNODE *z);

void walk_inorder(RBTNODE *tree);
void do_stuff(RBTNODE *tree, void *args);
void slow_range_query(RBTNODE* tree, void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, RBTNODE *node);
void rbtree_range_query(RBTNODE* tree, void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, RBTNODE *node);
void print_tree(RBTNODE* tree);
RBTNODE *rbtree_find_node(RBTNODE *tree, RBTNODE *node);
RBTNODE *rbtree_find_min(RBTNODE *tree);

#endif /* _PWNC_RBTREE_H */
